import os, logging, pathlib, datetime, pandas as pd, errno
from Python.Packages.caluxPy_bc.src.caluxPy_bc.ExtraccionBGD import ExtraccionBGD as Extraccion
from CalculoPosicion import CalculoPosicion as PNME

class ExtPosMid():
    
    def __init__(self, 
                 cwd: str, 
                 type: str, 
                 listado_cuentas: list[str], 
                 fecha: str = '', 
                 años: list = [], 
                 line_thresh: int = 3350, 
                 estructura: dict = {}, 
                 ext: bool = True, 
                 pos: bool = True,
                 file: str = '') -> None:
        
        self.cwd = cwd
        self.n_ext = 'prueba_pnme.xlsx'
        self.n_pos = 'resultados_pnme.xlsx'
        self.type = type
        self.fecha = fecha
        self.file = file
        
        self.crear_logger()
        
        #Verificación de que se incluyó hasta dentro de los parámetros       
        if self.type in ['update', 'big_update']:
            if self.fecha == '':
                e_prompt = 'Parámetro "fecha" no especificado'
                self.errorHandler(e_prompt)
            else:
                fecha_inputs = fecha.split('-')
        else:
            self.fecha = ''  
          
        self.años = años if self.fecha == '' else [fecha_inputs[0]]
        #self.años = ['2022']
        self.meses = ['ENE', 'FEB', 'MAR', 'ABR', 'MAY', 'JUN', 'JUL', 'AGO', 'SEP', 'OCT', 'NOV', 'DIC'] if self.fecha == '' else [self.mes_select(fecha_inputs[1])]
        #self.meses = ['FEB']
        self.dias = ['DIA' + str(1 + i) for i in range(31)] if self.fecha == '' else ['DIA' + str(int(fecha_inputs[2]))]
        self.mensuales = [mes + año[-2:] for año in años for mes in self.meses]
        self.columnas = ['AC', 'AB']
        
        #Asignación de otras variables
        self.estructura, self.listado_cuentas, self.line_thresh = estructura, listado_cuentas, line_thresh
        self.ext, self.pos = ext, pos
        
        #Verificación de cual proceso se desea
        if self.ext:
            self.datos = self.extraccion()
            self.exportar(datos = self.datos, df_type = 'extraccion')
        if self.pos:
            self.pnme = self.posicion()
    
        self.cerrar_logger()
        
    def extraccion (self) -> pd.DataFrame:
        df = pd.DataFrame()
        try:
            e_prompt = ''
            datos = Extraccion(estructura = self.estructura,
                               meses = self.meses,
                               años = self.años,
                               listado_cuentas = self.listado_cuentas,
                               dias = self.dias,
                               mensuales = self.mensuales,
                               columnas = self.columnas,
                               line_thresh = self.line_thresh,
                               cwd = self.cwd)
            df = datos.generated_df
            rutas_trabajadas = datos.rutas_trabajadas
            rutas_no_trabajadas = datos.rutas_no_trabajadas
            #print(f'Ruta -> {datos._ruta} \n Validación -> {datos._validacion}')
            #self._logger.info(f'Ruta -> {datos._ruta} \n Validación -> {datos._validacion}')
            #if datos.errores > 0:
            #    for errores in datos.errores_prompt:
            #        self._logger.info(f'{errores} || Mes: {mes} || Subsector: {subsector}')
            #rutas_trabajadas.append(datos._ruta)
        #except KeyError as e:
        #    e_prompt = f'--Error! -> Falta columna {e} en {mes}'
        #    self._logger.error(e_prompt)
        #except OSError as e:
        #    print(e)
        #    self._logger.error(e)
        #except IndexError as e:
        #    e_prompt = f'--Error! -> Archivo no existe ({e})'
        #    self._logger.error(e_prompt)
        #finally:
        #    if e_prompt != '': raise Exception(e_prompt)
        except Exception as e:
            self._logger.error(f'ExtPosMid_extraccion => {e}')
            
        #Si falla aquí es porque el codigo esta tomando un valor que no debería, por ejemplo una entidad en blanco
        df['Codigos'] = df['Subsector'] + ' ' + df['Entidad']
        df['Codigos'] = df['Codigos'].apply(lambda x: x.strip())
        df['Fecha'] = pd.to_datetime(df['Fecha'])  
        df = df.assign(Periodicidad=lambda x: x.apply(lambda row: 'Diaria' if row['Dia'][0:3] == 'DIA' else ('Mensual' if row['Mes'] == 'Mensual' else 'Other'), axis=1))
        
        #Mejorar esto
        rutas_todas = datos.rutas
        self._logger.info(f'Cantidad de Rutas Trabajadas: {len(rutas_trabajadas)}\nRutas Totales: {len(rutas_todas)}')
        print(f'Cantidad de Rutas Trabajadas: {len(rutas_trabajadas)}\nRutas Totales: {len(rutas_todas)}')
        if len(rutas_trabajadas) == len(rutas_todas):
            print('SUCCESS! -> Todas las rutas fueron trabajadas con éxito..')
            self._logger.info('SUCCESS! -> Todas las rutas fueron trabajadas con éxito..')
        else:
            rutas_no_trabajadas = []
            for ruta in rutas_todas:
                if not ruta in rutas_trabajadas:
                    rutas_no_trabajadas.append(ruta)
                    #print(rutas_no_trabajadas)
                    self._logger.info(f'Rutas no trabajadas => {rutas_no_trabajadas}')
       
        return df

    def posicion(self) -> pd.DataFrame:
        e_prompt = ''
        try:
            df = self.datos if self.ext == True else pd.read_excel(self.file)
            pnme = PNME(df, include = 'none', type = self.type, cwd = self.cwd)
        except ValueError as e:
            e_prompt = f'Revisar tipos de cambio => {e}'
            self._logger.error(e_prompt)
        except Exception as e:
            e_prompt = f'Error calculando posición => {e}'
            self._logger.error(e_prompt)
        finally:
            if e_prompt != '':
                raise Exception(e_prompt)
            else:
                return pnme

    def exportar(self, datos: pd.DataFrame, df_type: str) -> None:
        e_prompt = ''
        try:
            nombre = self.n_ext if df_type == 'extraccion' else self.n_pos if df_type == 'posicion' else None
            df = self.datos if df_type == 'extraccion' else self.pnme if df_type == 'posicion' else None
            if nombre == '' or df.empty:
                self._logger.error('Error iniciando exportado')
                raise Exception('Error iniciando exportado')
            ruta = os.path.join(self.cwd, nombre)
            if self.type == 'update':
                df_read = pd.read_excel(ruta)
                df_export = pd.concat([df_read, df])
                df_export.to_excel(ruta)
            elif self.type == 'new':
                df.to_excel(ruta)       
        except Exception as e:
            e_prompt = f'Error en exportación => {e}'
            self._logger.error(e_prompt)
        finally:
            if e_prompt != '': raise Exception(e_prompt)
        
    def mes_select(self, mes: str) -> str:
        if mes in ['01', '1']:
            return 'ENE'
        if mes in ['02', '2']:
            return 'FEB'
        if mes in ['03', '3']:
            return 'MAR'
        if mes in ['04', '4']:
            return 'ABR'
        if mes in ['05', '5']:
            return 'MAY'
        if mes in ['06', '6']:
            return 'JUN'
        if mes in ['07', '7']:
            return 'JUL'
        if mes in ['08', '8']:
            return 'AGO'
        if mes in ['09', '9']:
            return 'SEP'
        if mes in ['10', '10']:
            return 'OCT'
        if mes in ['11', '11']:
            return 'NOV'
        if mes in ['12', '12']:
            return 'DIC'

    def get_folder_path(self) -> str:
        # Get the path to the user's home directory
        home = pathlib.Path.home()

        # Construct the path to the desktop
        desktop = home / "Desktop"

        # Check if the desktop path exists; this is useful for OS's with different desktop paths
        if not desktop.exists():
            # Alternative approach for systems where the desktop might have a different name
            desktop = os.path.join(os.path.expanduser("~"), "Desktop")
        
        if not os.path.exists(desktop):
            folder_path = os.path.join(home, 'cxbcLogs')
        else:
            folder_path = os.path.join(desktop, 'cxbcLogs')
        
        if not os.path.exists(folder_path):
            os.makedirs(folder_path)
                
        return folder_path
    
    def crear_logger(self) -> None:
        #desktop = self.get_folder_path()
        _logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.DEBUG)
        self._fileHandler = logging.FileHandler("{0}/{1}.log".format(self.cwd, 'process_start'), mode = 'w', encoding = 'utf-8')
        self._fileHandler.setFormatter(_logFormatter)
        self._logger.addHandler(self._fileHandler)
        
    def cerrar_logger(self) -> None:
        self._logger.removeHandler(self._fileHandler)
        self._fileHandler.close()
    
    def errorHandler(self, prompt: str) -> None:
        self._logger.error(prompt)
        self.cerrar_logger()
        raise Exception(prompt)
    
    def crear_ruta(self, division, estructura, subsector, año, mes):
        terminal = ['.xlsm', '.xlsx']
        terminal_seeker = 0
        try:
            while True:
                if division == 'Division_Banca_Fomento':
                    directorio = estructura[division]['estructura_ruta'].format(division = division, año = año, subsector = subsector)
                    archivo = estructura[division]['estructura_archivo'].format(subsector = subsector, mes = mes, año = año[-2:], terminal = terminal[terminal_seeker])
                elif division == 'Division_Banca_Comercial':
                    directorio = estructura[division]['estructura_ruta'].format(division = division, año = año)
                    archivo = estructura[division]['estructura_archivo'].format(mes = mes, año = año[-2:], terminal = terminal[terminal_seeker])
                ruta = os.path.join(directorio, archivo)
                if os.path.exists(ruta):
                    break
                else:
                    terminal_seeker += 1
                    if terminal_seeker >= len(terminal): raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), ruta) 
        except Exception as e:
            self._logger.error(f'{e} => {ruta}')
        
        return ruta


import os, pandas as pd, datetime, errno, logging, traceback
from concurrent.futures import ThreadPoolExecutor, as_completed

class ExtraccionBGD():
   
    def __init__(self, estructura, meses, años, listado_cuentas, dias, mensuales, columnas, line_thresh, cwd) -> None:
        
        #Definición de las Variables de la Instancia
        self.estructura = estructura
        self.meses = meses
        self.años = años
        self.listado_cuentas = listado_cuentas
        self.dias = dias
        self.mensuales = mensuales
        self.columnas = columnas
        self.errores = 0
        self.errores_prompt = ''
        self.line_thresh = line_thresh
        self.cwd = cwd
        self.rutas_trabajadas = []
        self.rutas_no_trabajadas = []
        
        self.crear_logger()
        hoy = datetime.datetime.now().date()
        
        #Ejecución de las Funcione
        self.rutas = []
        for año in self.años:
            conteo_mes = 0
            for mes in meses:
                conteo_mes += 1
                if int(año) >= hoy.year and conteo_mes > hoy.month:
                    break
                for key, value in estructura.items():
                    for subsector in value['subsectores']:
                        self.rutas.append(self.crear_ruta(estructura = estructura, division = key, subsector = subsector, año = año, mes = mes))

        self.generated_df = self.recoleccion_de_datos_concurrent(rutas = self.rutas, 
                                                                 listado_cuentas = self.listado_cuentas, 
                                                                 dias = self.dias, 
                                                                 mensuales = self.mensuales, 
                                                                 columnas = self.columnas, 
                                                                 line_thresh = self.line_thresh)
        self.cerrar_logger()
        
    def crear_ruta(self, division, estructura, subsector, año, mes):
        terminal = ['.xlsm', '.xlsx']
        terminal_seeker = 0
        try:
            while True:
                if division == 'Division_Banca_Fomento':
                    directorio = estructura[division]['estructura_ruta'].format(division = division, año = año, subsector = subsector)
                    archivo = estructura[division]['estructura_archivo'].format(subsector = subsector, mes = mes, año = año[-2:], terminal = terminal[terminal_seeker])
                elif division == 'Division_Banca_Comercial':
                    directorio = estructura[division]['estructura_ruta'].format(division = division, año = año)
                    archivo = estructura[division]['estructura_archivo'].format(mes = mes, año = año[-2:], terminal = terminal[terminal_seeker])
                ruta = os.path.join(directorio, archivo)
                if os.path.exists(ruta):
                    break
                else:
                    terminal_seeker += 1
                    if terminal_seeker >= len(terminal): raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), ruta) 
        except Exception as e:
            self._logger.error(f'crear_ruta => {e} => {ruta}')
        
        return ruta

    def recoleccion_de_datos_concurrent(self, rutas, listado_cuentas, dias, mensuales, columnas, line_thresh):
        # Initialize DataFrame to collect all results
        data_frames = []
        all_data = pd.DataFrame()
        # Use ThreadPoolExecutor to process multiple files in parallel
        
        max_workers = min(10, os.cpu_count() + 4) 
        with ThreadPoolExecutor(max_workers = max_workers) as executor:
            # Map each file to a future
            futures = {executor.submit(self.process_file, ruta, listado_cuentas, dias, mensuales, columnas, line_thresh): ruta for ruta in rutas}
            for future in as_completed(futures):
                try:
                    file_data = future.result()
                    data_frames.append(file_data)
                except Exception as e:
                    self._logger.error(f"recoleccion_de_datos_concurrent => Error processing file {futures[future]}: {e}")
                    print(f"recoleccion_de_datos_concurrent => Error processing file {futures[future]}: {e}")
        all_data = pd.concat(data_frames, ignore_index = True)
        if all_data.empty:
            raise ValueError('recoleccion_de_datos_concurrent => No data collected.')
        else:
            self.limpiar_data(all_data)
        return all_data

    def process_file(self, file_path, listado_cuentas, dias, mensuales, columnas, line_thresh):
        # Initialize a container for DataFrames generated from each sheet
        data_frames = []
        processed_data = pd.DataFrame()
        # Use pandas.ExcelFile for efficient access to sheets
        with pd.ExcelFile(file_path) as excel:
            print(f'Trabajando => {file_path}')
            try:
                for sheet_name in excel.sheet_names:
                    if sheet_name in dias + mensuales:
                        col = 0
                        while True:
                            try:
                                df = excel.parse(sheet_name = sheet_name, usecols = f'A:{columnas[col]}', nrows = line_thresh)
                                processed_df, status, error = self.generar_df_concurrent(df, file_path, sheet_name, listado_cuentas)
                                data_frames.append(processed_df)
                                break
                            except pd.errors.ParserError as e:
                                col += 1
                                if col >= len(columnas): raise pd.errors.ParserError('Column could not be properly parsed')
                            except Exception as e:
                                error_info = traceback.format_exc()
                                raise Exception(f'{e}\n{error_info}')
                        if processed_df.empty:
                            if status == 'no trabajado':
                                print(f'\tprocess_file => from {file_path}: {sheet_name} status: {status}')
                            else:
                                print(f'\tprocess_file => from {file_path}: {sheet_name} => empty')
                                
            except Exception as e:
                error_info = traceback.format_exc()
                raise Exception(f'process_file => {e}\n{error_info}')
            processed_data = pd.concat(data_frames, ignore_index=True)
            if processed_data.empty:
                print(f'process_file => file {file_path} did not return data')
                self._logger.info(f'process_file => file {file_path} did not return data')
                self.rutas_no_trabajadas.append(file_path)
            else:
                print(f'Trabajada => {file_path}')
                self.rutas_trabajadas.append(file_path)
        # Concatenate all DataFrames from this file
        return processed_data

    def generar_df_concurrent(self, df, file_path, sheet_name, listado_cuentas) -> list[pd.DataFrame, str]:
        processed_data_df = pd.DataFrame()
        # Load the specific sheet into a DataFrame
        try:
            # Assuming `extraer_data` can process data from a single DataFrame
            processed_data, tipo_cambio, fecha, status, error = self.extraer_data(data = df, listado_cuentas = listado_cuentas, file_name = os.path.basename(file_path))
            if error != '':
                if len(processed_data) > 0:
                    # Additional processing, e.g., adding columns for Dia, Mes, etc.
                    processed_data['Dia'] = sheet_name  # Assuming sheet names are dates or similar
                    processed_data['Mes'], processed_data['Subsector'] = self.get_splitted(file_path)
                    processed_data['Tipo Cambio'] = tipo_cambio
                    processed_data['Fecha'] = fecha
                    #processed_data['Periodicidad'] = 'Diaria' if sheet_name[0:3] == 'DIA' else 'Mensual'
                    
                    # Reorder or modify DataFrame as needed
                    processed_data_df = pd.DataFrame(processed_data)
                    processed_data_df = processed_data_df[['Dia', 'Mes', 'Fecha', 'Subsector', 'Entidad', 'Cuenta', 'Valor', 'Tipo Cambio']]
                    if processed_data_df.empty:
                        print(f'generar_df_concurrent => {file_path}: {sheet_name} (empty)')
        except KeyError as e:
            error_info = traceback.format_exc()
            self._logger.error(f'generar_df_concurrent => Error Falta columna {e} || Día: {sheet_name}\n{error_info}')
        except Exception as e:
            error_info = traceback.format_exc()
            self._logger.error(f"generar_df_concurrent => Error processing sheet {sheet_name}: {file_path} => {e}\n{error_info}")
            
        return processed_data_df, status, error  # Return an empty DataFrame in case of error

    def extraer_data(self, data, listado_cuentas, file_name) -> list[dict, float, datetime.date, str]:
        valores_data = []
        entidades_data = []
        cuentas_data = []
        datos = {}
        tipo_cambio = 0
        fecha = None
        error = 0
        status = 'ok'
        
        try:
            fecha = data.iloc[0].index[0]
            if fecha in [None, '', 'Unnamed: 0']:
                status = 'no trabajado'
                raise ValueError(f'No trabajado')
            tipo_cambio = float(data.iloc[0].iloc[1]) #Extracción del tipo de cambio
            hoy = datetime.datetime.now().date()
            entidades = data.iloc[2].to_list() #Extracción del nombre de las entidades
            entidades[0] = 'ENTIDADES' #reemplaza el primer valor por esto
            data.rename(columns = data.iloc[2], inplace = True) #Cambia la primera linea del DataFrame para que los índices por columna sean adecuados
            #data.columns = entidades

            #Definición de Variables
            cuenta = ''
            if isinstance(fecha, datetime.datetime):
                fecha = fecha.date()
                if fecha <= hoy:
                    #Proceso de creado de listas
                    for codigo in listado_cuentas:
                        try:
                            linea = data[data['CODIGOS'] == codigo].values[0].tolist()
                            cuenta = linea[0]
                            for item in linea: #itera sobre las cuentas seleccionadas
                                #if type(item) == int or type(item) == float and not pd.isna(item): #itera por cada componente de la línea y si cumple con las condiciones se agrega
                                if item != codigo and item != cuenta:
                                    valores_data.append(item) #Se agrega el valor a los datos
                                    cuentas_data.append(f'{cuenta.lstrip(' ').rstrip(' ').title()} ({codigo})') #Se agrega la cuenta por cada dato agregado
                                    
                            for entidad in entidades:
                                if entidad != 'ENTIDADES' and entidad != 'CODIGOS': #and not pd.isna(entidad) and entidad != 0:
                                    entidades_data.append(entidad)

                            #Verifica que las longitudes cuadren y si no devuelve un error
                            if len(entidades_data) == len(valores_data):
                                datos = {
                                        'Entidad': entidades_data,
                                        'Cuenta': cuentas_data,
                                        'Valor': valores_data
                                        }
                        except Exception as e: #Si no encuentra la cuenta
                            error_info = traceback.format_exc()
                            self._logger.error(f'extraer_data => ERROR Cuenta {cuenta}: {codigo} no encontrada en fecha {fecha} => {file_name} \n{error_info}')
        
        except ValueError as e:
            pass
        except Exception as e:
            error_info = traceback.format_exc()
            error = f'extraer_data => {e}\n {error_info}'
        finally:
            return datos, tipo_cambio, fecha, status, error

    def limpiar_data(self, df):
        df.dropna(inplace = True)

    def get_splitted(self, ruta: str) -> str:
            splitted = ruta.split('\\')
            mes = splitted[-1].split('.')[0].split('_')[-1][0:3]
            if splitted[5] == 'Division_Banca_Fomento':
                subsector = splitted[8]
            elif splitted[5] == 'Division_Banca_Comercial':
                subsector = 'BM'
            return mes, subsector
        
    def crear_logger(self) -> None:
        #desktop = self.get_folder_path()
        _logFormatter = logging.Formatter("%(asctime)s: [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s")
        self._logger = logging.getLogger()
        self._logger.setLevel(logging.DEBUG)
        self._fileHandler = logging.FileHandler("{0}/{1}.log".format(self.cwd, 'extraccion_log'), mode = 'w', encoding = 'utf-8')
        self._fileHandler.setFormatter(_logFormatter)
        self._logger.addHandler(self._fileHandler)
        
    def cerrar_logger(self) -> None:
        self._logger.removeHandler(self._fileHandler)
        self._fileHandler.close()
  
    #region => Backup
    '''
    def recoleccion_de_datos(self, excel, listado_cuentas, dias, mensuales, columnas, subsector, mes, line_thresh):      
        #Llama la función y como se manda el DataFrame entonces se concatena la información al usado
        generated_df = self.generar_df(
                                        excel = excel, 
                                        listado_cuentas = listado_cuentas, 
                                        dias = dias, 
                                        mensuales = mensuales,
                                        columnas = columnas,
                                        subsector = subsector,
                                        mes = mes,
                                        line_thresh = line_thresh
                                        )
        del excel
        return generated_df
    
    def generar_df(self, excel, listado_cuentas, dias, mensuales, columnas, subsector, mes, line_thresh):
        #Genera una base concatenada de los dias y ese DF generado lo devuelve a recolección de datos
        generated_df = pd.DataFrame()
        self.errores_prompt = []
        
        for dia in excel.sheet_names:
            if dia in dias or dia in mensuales:
                col = 0
                while True:
                    try:
                        data, tipo_cambio, fecha = self.extraer_data(
                                                                     data = pd.read_excel(excel, sheet_name = dia, nrows = line_thresh, usecols = columnas[col]), 
                                                                     listado_cuentas = listado_cuentas
                                                                    )
                        data = pd.DataFrame(data)
                        #Agrega columnas de constantes
                        if not data.empty:
                            data['Dia'] = dia
                            data['Mes'] = mes
                            data['Subsector'] = subsector
                            data['Tipo Cambio'] = tipo_cambio
                            data['Fecha'] = fecha
                            data = data[['Dia', 'Mes', 'Fecha', 'Subsector', 'Entidad', 'Cuenta', 'Valor', 'Tipo Cambio']]
                            
                        #Concatena las tablas que se van creando
                        generated_df = pd.concat([generated_df, data])
                        break
                    except pd.errors.ParserError as e:
                        col += 1
                    except KeyError as e:
                        message = f'--Error! -> Falta columna {e} || Día: {dia}'
                        self._logger.error(message)
                        self.errores += 1
                        self.errores_prompt.append(f' {message}')
                        raise KeyError(self.errores_prompt)
                        
        return generated_df
    
    def verificar_ruta(self, ruta):
        if os.path.exists(ruta):
            return True
        else: 
            return False
    
    def crear_ruta(self, division, estructura, subsector, año, mes):
        terminal_seeker = 0
        ruta = ''
        try:
            while True:
                if division == 'Division_Banca_Fomento':
                    directorio = estructura['estructura_ruta'].format(division = division, año = año, subsector = subsector)
                    #Se usa terminal seeker para iterar en opciones
                    archivo = estructura['estructura_archivo'].format(subsector = subsector, mes = mes, año = año[-2:], terminal = terminal[terminal_seeker])
                else: 
                    directorio = estructura['estructura_ruta'].format(division = division, año = año)
                    archivo = estructura['estructura_archivo'].format(mes = mes, año = año[-2:], terminal = terminal[terminal_seeker])
                ruta = os.path.join(directorio, archivo)
                validacion = self.verificar_ruta(ruta = ruta)
                if validacion == True:
                    break
                else:
                    terminal_seeker += 1
                    if terminal_seeker > 1: raise Exception('Ruta no encontrada')
            ExtraccionBGD.rutas.append(ruta)
        except Exception as e:
            self._logger.error(f'{e} => {ruta}')
            raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), ruta) 
            
        return ruta, validacion'''
    #endregion => Backup




import os
import pandas as pd
import datetime
import logging

class ExtraccionBGD:

    rutas = []

    def __init__(self, estructura, division, mes, año, listado_cuentas, dias, mensuales, columnas, subsector, line_thresh, cwd, fecha=None):
        """
        Initialize the ExtraccionBGD instance with provided parameters.
        """
        self.estructura = estructura
        self.division = division
        self.mes = mes
        self.año = año
        self.listado_cuentas = listado_cuentas
        self.dias = dias
        self.mensuales = mensuales
        self.columnas = columnas
        self.subsector = subsector
        self.terminal = ['.xlsm', '.xlsx']
        self.errores = 0
        self.errores_prompt = []
        self.line_thresh = line_thresh
        self.cwd = cwd
        self.Fecha = None  # Default value if no fecha is provided
        if fecha:
            self.Fecha = datetime.datetime.strptime(fecha, '%Y-%m-%d').date()

        logging.basicConfig(filename=os.path.join(self.cwd, 'extraccion_log.txt'), level=logging.DEBUG,
                            format='%(asctime)s - %(levelname)s - %(message)s')

        self._ruta, self._validacion = self.crear_ruta()
        self.generated_df = self.recoleccion_de_datos()
        self.limpiar_data(self.generated_df)

    def crear_ruta(self):
        """
        Create the file path based on the given parameters and verify its existence.
        """
        terminal_seeker = 0
        ruta = ''

        while True:
            if self.division == 'Division_Banca_Fomento':
                directorio = self.estructura['estructura_ruta'].format(
                    division=self.division,
                    año=self.año,
                    subsector=self.subsector
                )
                archivo = self.estructura['estructura_archivo'].format(
                    subsector=self.subsector,
                    mes=self.mes,
                    año=self.año[-2:],
                    terminal=self.terminal[terminal_seeker]
                )
            else:
                directorio = self.estructura['estructura_ruta'].format(
                    division=self.division,
                    año=self.año
                )
                archivo = self.estructura['estructura_archivo'].format(
                    mes=self.mes,
                    año=self.año[-2:],
                    terminal=self.terminal[terminal_seeker]
                )

            ruta = os.path.join(directorio, archivo)
            validacion = self.verificar_ruta(ruta)
            if validacion:
                break
            else:
                terminal_seeker += 1
                if terminal_seeker >= len(self.terminal):
                    logging.error(f"No valid file found for {self.division}, {self.subsector}, {self.mes}, {self.año}.")
                    raise FileNotFoundError("No valid file found.")
                    
        ExtraccionBGD.rutas.append(ruta)
        return ruta, validacion

    @staticmethod
    def verificar_ruta(ruta):
        """
        Verify if the given file path exists.
        """
        if os.path.exists(ruta):
            logging.info(f"File found: {ruta}")
            return True
        else:
            logging.warning(f"File not found: {ruta}")
            return False

    def recoleccion_de_datos(self):
        """
        Collect data from the specified Excel file.
        """
        excel = pd.ExcelFile(self._ruta)
        generated_df = self.generar_df(excel)
        del excel
        return generated_df

    def generar_df(self, excel):
        """
        Generate a DataFrame by concatenating data from the specified sheets.
        """
        generated_df = pd.DataFrame()

        for dia in excel.sheet_names:
            if dia in self.dias or (dia in self.mensuales and (not self.Fecha or self.check_fecha_match(dia))):
                col = 0
                while True:
                    try:
                        data, tipo_cambio, fecha = self.extraer_data(
                            pd.read_excel(excel, sheet_name=dia, nrows=self.line_thresh, usecols=self.columnas[col])
                        )
                        if self.Fecha and fecha != self.Fecha:
                            break  # Exit if the extracted Fecha does not match the class variable Fecha

                        data = pd.DataFrame(data)
                        if not data.empty:
                            data['Dia'] = dia
                            data['Mes'] = self.mes
                            data['Subsector'] = self.subsector
                            data['Tipo Cambio'] = tipo_cambio
                            data['Fecha'] = fecha
                            # Create a unique identifier for each row
                            data['Unique_ID'] = data.apply(
                                lambda row: f"{row['Dia']}_{row['Mes']}_{row['Subsector']}_{row['Entidad']}_{row['Cuenta']}_{row['Fecha']}", 
                                axis=1
                            )
                            data = data[['Unique_ID', 'Dia', 'Mes', 'Fecha', 'Subsector', 'Entidad', 'Cuenta', 'Valor', 'Tipo Cambio']]
                        
                        generated_df = pd.concat([generated_df, data])
                        break
                    except pd.errors.ParserError:
                        col += 1
                        if col >= len(self.columnas):
                            logging.error(f"Parser error in sheet {dia} for {self.subsector}, {self.mes}, {self.año}.")
                            break
                    except KeyError as e:
                        message = f'--Error! -> Falta columna {e} || Día: {dia}'
                        self.errores += 1
                        self.errores_prompt.append(message)
                        logging.error(message)
                        break

        return generated_df

    def check_fecha_match(self, dia):
        """
        Check if the fecha extracted matches the class variable Fecha.
        """
        if not self.Fecha:
            return True
        
        excel = pd.ExcelFile(self._ruta)
        for col in self.columnas:
            try:
                _, _, fecha = self.extraer_data(
                    pd.read_excel(excel, sheet_name=dia, nrows=self.line_thresh, usecols=col)
                )
                if fecha == self.Fecha:
                    return True
            except pd.errors.ParserError:
                continue
            except KeyError:
                continue
        return False

    def extraer_data(self, data):
        """
        Extract data from a DataFrame and return the processed values.
        """
        tipo_cambio = data.iloc[0].iloc[1]
        fecha = data.iloc[0].index[0]
        hoy = datetime.datetime.now().date()
        entidades = data.iloc[2].to_list()
        entidades[0] = 'ENTIDADES'
        data.rename(columns=data.iloc[2], inplace=True)

        valores_data = []
        entidades_data = []
        cuentas_data = []
        datos = {}

        if isinstance(fecha, datetime.datetime):
            fecha = fecha.date()
            if fecha <= hoy:
                for codigo in self.listado_cuentas:
                    try:
                        linea = data[data['CODIGOS'] == codigo].values[0].tolist()
                        cuenta = linea[0]
                        for item in linea:
                            if item != codigo and item != cuenta:
                                valores_data.append(item)
                                cuentas_data.append(f'{cuenta.strip().title()} ({codigo})')

                        for entidad in entidades:
                            if entidad != 'ENTIDADES' and entidad != 'CODIGOS':
                                entidades_data.append(entidad)

                        if len(entidades_data) == len(valores_data):
                            datos = {
                                'Entidad': entidades_data,
                                'Cuenta': cuentas_data,
                                'Valor': valores_data
                            }
                        else:
                            logging.error(f"Length mismatch between entidades and valores for {codigo} on {fecha}.")
                            logging.error(f"Entidades: {entidades_data}")
                            logging.error(f"Valores: {valores_data}")
                    except Exception as e:
                        logging.error(f"ERROR! --> Cuenta {cuenta} no encontrada: {e}")

        return datos, tipo_cambio, fecha

    @staticmethod
    def limpiar_data(df):
        """
        Clean the DataFrame by dropping NaN values and removing duplicates.
        """
        df.dropna(inplace=True)
        initial_len = len(df)
        df.drop_duplicates(subset=['Unique_ID'], inplace=True)
        final_len = len(df)
        
        logging.info(f"Removed {initial_len - final_len} duplicate rows based on Unique_ID.")
        
        # Drop the 'Unique_ID' column
        df.drop(columns=['Unique_ID'], inplace=True)
        logging.info("Dropped 'Unique_ID' column after removing duplicates.")




import os
import pandas as pd
import datetime
import logging

class ExtraccionBGD:

    TERMINALS = ['.xlsm', '.xlsx']
    LOG_FILE = 'extraccion_log.txt'
    
    def __init__(self, estructura, division, mes, año, listado_cuentas, dias, mensuales, columnas, subsector, line_thresh, cwd):
        """
        Initialize the ExtraccionBGD instance with provided parameters.
        """
        self.estructura = estructura
        self.division = division
        self.mes = mes
        self.año = año
        self.listado_cuentas = listado_cuentas
        self.dias = dias
        self.mensuales = mensuales
        self.columnas = columnas
        self.subsector = subsector
        self.terminal = ['.xlsm', '.xlsx']
        self.errores = 0
        self.errores_prompt = []
        self.line_thresh = line_thresh
        self.cwd = cwd
        #self.Fecha = datetime.datetime.strptime(fecha, '%Y-%m-%d').date()  # Add the new class variable Fecha
        self.rutas = []

        log_path = os.path.join(self.cwd, self.LOG_FILE)
        logging.basicConfig(
            filename=log_path,
            level=logging.DEBUG,
            format='%(asctime)s - %(levelname)s - %(message)s'
        )

        self._ruta, self._validacion = self.crear_ruta()
        self.generated_df = self.recoleccion_de_datos()
        self.limpiar_data(self.generated_df)

    @staticmethod
    def validate_fecha(fecha):
        """
        Validate and convert the input date string to a datetime.date object.
        """
        try:
            return datetime.datetime.strptime(fecha, '%Y-%m-%d').date()
        except ValueError:
            logging.error("Invalid date format. Expected 'YYYY-MM-DD'.")
            raise
  
    def crear_ruta(self):
        """
        Create the file path based on the given parameters and verify its existence.
        """
        for terminal in self.TERMINALS:
            directorio = self.estructura['estructura_ruta'].format(
                division = self.division,
                año = self.año,
                subsector = getattr(self, 'subsector', '')
            )
            archivo = self.estructura['estructura_archivo'].format(
                subsector=self.subsector if self.division == 'Division_Banca_Fomento' else '',
                mes=self.mes,
                año=self.año[-2:],
                terminal=terminal
            )
            ruta = os.path.join(directorio, archivo)
            if self.verificar_ruta(ruta):
                self.rutas.append(ruta)
                return ruta, True
        '''
        while True:
            if self.division == 'Division_Banca_Fomento':
                directorio = self.estructura['estructura_ruta'].format(
                    division=self.division,
                    año=self.año,
                    subsector=self.subsector
                )
                archivo = self.estructura['estructura_archivo'].format(
                    subsector=self.subsector,
                    mes=self.mes,
                    año=self.año[-2:],
                    terminal=self.terminal[terminal_seeker]
                )
            else:
                directorio = self.estructura['estructura_ruta'].format(
                    division=self.division,
                    año=self.año
                )
                archivo = self.estructura['estructura_archivo'].format(
                    mes=self.mes,
                    año=self.año[-2:],
                    terminal=self.terminal[terminal_seeker]
                )

            ruta = os.path.join(directorio, archivo)
            validacion = self.verificar_ruta(ruta)
            if validacion:
                break
            else:
                terminal_seeker += 1
                if terminal_seeker >= len(self.terminal):
                    logging.error(f"No valid file found for {self.division}, {self.subsector}, {self.mes}, {self.año}.")
                    raise FileNotFoundError("No valid file found.")
                    
        ExtraccionBGD.rutas.append(ruta)
        return ruta, validacion
        '''
        logging.error("No valid file found for the specified parameters.")
        raise FileNotFoundError("No valid file found.")
  
    @staticmethod
    def verificar_ruta(ruta):
        """
        Verify if the given file path exists.
        """
        if os.path.exists(ruta):
            logging.info(f"File found: {ruta}")
            return True
        else:
            logging.warning(f"File not found: {ruta}")
            return False

    def recoleccion_de_datos(self):
        """
        Collect data from the specified Excel file.
        """
        '''
        excel = pd.ExcelFile(self._ruta)
        generated_df = self.generar_df(excel)
        del excel
        return generated_df
        '''
        
        with pd.ExcelFile(self.ruta) as excel:
            relevant_sheets = self.get_relevant_sheets(excel.sheet_names)
            return self.generar_df(excel, relevant_sheets)

    def get_relevant_sheets(self, sheet_names):
        """
        Filter and return relevant sheets based on `dias` and `mensuales`.
        """
        return [
            sheet for sheet in sheet_names
            if sheet in self.dias or (sheet in self.mensuales) #and self.check_fecha_match(sheet))
        ]

    def generar_df(self, excel, sheets):
        """
        Generate a DataFrame by concatenating data from the specified sheets.
        """
        generated_df = pd.DataFrame()

        '''
        for dia in excel.sheet_names:
            if dia in self.dias or (dia in self.mensuales and self.check_fecha_match(dia)):
                col = 0
                while True:
                    try:
                        data, tipo_cambio, fecha = self.extraer_data(
                            pd.read_excel(excel, sheet_name=dia, nrows=self.line_thresh, usecols=self.columnas[col])
                        )
                        if self.Fecha and fecha != self.Fecha:
                            break  # Exit if the extracted Fecha does not match the class variable Fecha

                        data = pd.DataFrame(data)
                        if not data.empty:
                            data['Dia'] = dia
                            data['Mes'] = self.mes
                            data['Subsector'] = self.subsector
                            data['Tipo Cambio'] = tipo_cambio
                            data['Fecha'] = fecha
                            # Create a unique identifier for each row
                            data['Unique_ID'] = data.apply(lambda row: f"{row['Dia']}_{row['Mes']}_{row['Subsector']}_{row['Entidad']}_{row['Cuenta']}_{row['Fecha']}", axis=1)
                            data = data[['Unique_ID', 'Dia', 'Mes', 'Fecha', 'Subsector', 'Entidad', 'Cuenta', 'Valor', 'Tipo Cambio']]
                        
                        generated_df = pd.concat([generated_df, data])
                        break
                    except pd.errors.ParserError:
                        col += 1
                        if col >= len(self.columnas):
                            logging.error(f"Parser error in sheet {dia} for {self.subsector}, {self.mes}, {self.año}.")
                            break
                    except KeyError as e:
                        message = f'--Error! -> Falta columna {e} || Día: {dia}'
                        self.errores += 1
                        self.errores_prompt.append(message)
                        logging.error(message)
                        break
        '''
        
        for sheet in sheets:
            try:
                for col in self.columnas:
                    try:
                        data, tipo_cambio, fecha = self.extraer_data(pd.read_excel(excel, sheet_name=sheet, nrows=self.line_thresh, usecols=col))
                        if not fecha:
                            logging.info(f"Skipping sheet {sheet} as fecha is not found.")
                            break
                        data = self.process_data(data, sheet, tipo_cambio, fecha)
                        generated_df = pd.concat([generated_df, data], ignore_index=True)
                        break
                    except pd.errors.ParserError:
                        logging.warning(f"Parser error in sheet {sheet}, column {col}.")
            except Exception as e:
                logging.exception(f"Error processing sheet {sheet}: {e}")
        
        return generated_df

    def process_data(self, data, sheet, tipo_cambio, fecha):
        """
        Process raw data into a structured DataFrame with metadata columns.
        """
        data['Dia'] = sheet
        data['Mes'] = self.mes
        data['Subsector'] = self.subsector
        data['Tipo Cambio'] = tipo_cambio
        data['Fecha'] = fecha
        data['Unique_ID'] = data.apply(
            lambda row: f"{row['Dia']}_{row['Mes']}_{row['Subsector']}_{row['Entidad']}_{row['Cuenta']}_{row['Fecha']}",
            axis=1
        )
        return data[['Unique_ID', 'Dia', 'Mes', 'Fecha', 'Subsector', 'Entidad', 'Cuenta', 'Valor', 'Tipo Cambio']]
        
    def check_fecha_match(self, dia):
        """
        Check if the fecha extracted matches the class variable Fecha.
        """
        excel = pd.ExcelFile(self._ruta)
        for col in self.columnas:
            try:
                _, _, fecha = self.extraer_data(
                    pd.read_excel(excel, sheet_name=dia, nrows=self.line_thresh, usecols=col)
                )
                if fecha == self.Fecha:
                    return True
            except pd.errors.ParserError:
                continue
            except KeyError:
                continue
        return False

    def extraer_data(self, data):
        """
        Extract data from a DataFrame and return the processed values.
        """
        try:
            tipo_cambio = data.iloc[0, 1] #tipo_cambio = data.iloc[0].iloc[1]
            fecha = data.columns[0] #fecha = data.iloc[0].index[0]
            hoy = datetime.datetime.now().date()
            entidades = data.iloc[2].to_list()
            entidades[0] = 'ENTIDADES'
            data.rename(columns=data.iloc[2], inplace=True)

            valores_data = []
            entidades_data = []
            cuentas_data = []
            datos = {}

            if isinstance(fecha, datetime.datetime):
                fecha = fecha.date()
                if fecha <= hoy:
                    for codigo in self.listado_cuentas:
                        try:
                            linea = data[data['CODIGOS'] == codigo].values[0].tolist()
                            cuenta = linea[0]
                            for item in linea:
                                if item != codigo and item != cuenta:
                                    valores_data.append(item)
                                    cuentas_data.append(f'{cuenta.strip().title()} ({codigo})')

                            for entidad in entidades:
                                if entidad != 'ENTIDADES' and entidad != 'CODIGOS':
                                    entidades_data.append(entidad)

                            if len(entidades_data) == len(valores_data):
                                datos = {
                                    'Entidad': entidades_data,
                                    'Cuenta': cuentas_data,
                                    'Valor': valores_data
                                }
                            else:
                                logging.error(f"Length mismatch between entidades and valores for {codigo} on {fecha}.")
                                logging.error(f"Entidades: {entidades_data}")
                                logging.error(f"Valores: {valores_data}")
                        except Exception as e:
                            logging.error(f"ERROR! --> Cuenta {cuenta} no encontrada: {e}")

            return datos, tipo_cambio, fecha
        except Exception as e:
            logging.exception(f"Error extracting data: {e}")
            return pd.DataFrame(), None, None

    @staticmethod
    def limpiar_data(df):
        """
        Clean the DataFrame by dropping NaN values and removing duplicates.
        """
        df.dropna(inplace=True)
        initial_len = len(df)
        df.drop_duplicates(subset=['Unique_ID'], inplace=True)
        final_len = len(df)
        
        logging.info(f"Removed {initial_len - final_len} duplicate rows based on Unique_ID.")
        
        # Drop the 'Unique_ID' column
        df.drop(columns=['Unique_ID'], inplace=True)
        logging.info("Dropped 'Unique_ID' column after removing duplicates.")




